@@ -37,7 +37,7 @@ import qualified Data.Text.Lazy.IO as TIO
 import Control.Monad (when)
 
 newtype Ident = Ident String
-    deriving (Show, Eq, Read, Data, Typeable)
+    deriving (Show, Eq, Read, Data, Typeable, Ord)
 
 type Scope = [(Ident, Exp)]
 
@@ -48,7 +48,8 @@ data Deref = DerefModulesIdent [String] Ident
            | DerefString String
            | DerefBranch Deref Deref
            | DerefList [Deref]
-    deriving (Show, Eq, Read, Data, Typeable)
+           | DerefTuple [Deref]
+    deriving (Show, Eq, Read, Data, Typeable, Ord)
 
 instance Lift Ident where
     lift (Ident s) = [|Ident|] `appE` lift s
@@ -76,16 +77,24 @@ instance Lift Deref where
         return $ dr `AppE` InfixE (Just n) per (Just d)
     lift (DerefString s) = [|DerefString|] `appE` lift s
     lift (DerefList x) = [|DerefList $(lift x)|]
+    lift (DerefTuple x) = [|DerefTuple $(lift x)|]
 
 derefParens, derefCurlyBrackets :: Parser Deref
 derefParens        = between (char '(') (char ')') parseDeref
 derefCurlyBrackets = between (char '{') (char '}') parseDeref
 
-derefList :: Parser Deref
+derefList, derefTuple :: Parser Deref
 derefList = between (char '[') (char ']') (fmap DerefList $ sepBy parseDeref (char ','))
+derefTuple = try $ do
+  _ <- char '('
+  x <- sepBy1 parseDeref (char ',')
+  when (length x < 2) $ pzero
+  _ <- char ')'
+  return $ DerefTuple x
 
 parseDeref :: Parser Deref
-parseDeref = skipMany (oneOf " \t") >> (derefList <|> (do
+parseDeref = skipMany (oneOf " \t") >> (derefList <|>
+                                        derefTuple <|> (do
     x <- derefSingle
     (derefInfix x) <|> (do
         res <- deref' $ (:) x
@@ -95,21 +104,20 @@ parseDeref = skipMany (oneOf " \t") >> (derefList <|> (do
     delim = (many1 (char ' ') >> return())
             <|> lookAhead (oneOf "(\"" >> return ())
     derefOp = try $ do
-        _ <- char '('
-        x <- many1 $ noneOf " \t\n\r()"
-        _ <- char ')'
-        return $ DerefIdent $ Ident x
+            _ <- char '('
+            x <- many1 $ noneOf " \t\n\r()"
+            _ <- char ')'
+            return $ DerefIdent $ Ident x
     derefInfix x = try $ do
-        () <- fail "Infix operator handling is currently disabled"
-        _ <- many1 $ oneOf " \t"
+        _ <- delim
+        xs <- many $ try $ derefSingle >>= \x' -> delim >> return x'
         op <- many1 (satisfy $ \c -> isSymbol c || c `elem` "-") <?> "operator"
         -- special handling for $, which we don't deal with
         when (op == "$") $ fail "don't handle $"
         let op' = DerefIdent $ Ident op
-        _ <- many1 (oneOf " \t")
-        y <- derefSingle
-        return $ DerefBranch (DerefBranch op' x) y
-    derefSingle = derefOp <|> derefParens <|> numeric <|> strLit<|> ident
+        ys <- many1 $ delim >> derefSingle
+        return $ DerefBranch (DerefBranch op' $ foldl1 DerefBranch $ x : xs) (foldl1 DerefBranch ys)
+    derefSingle = derefTuple <|> derefOp <|> derefParens <|> numeric <|> strLit<|> ident
     deref' lhs =
         dollar <|> derefSingle'
                <|> return (foldl1 DerefBranch $ lhs [])
@@ -176,6 +184,7 @@ derefToExp _ (DerefIntegral i) = LitE $ IntegerL i
 derefToExp _ (DerefRational r) = LitE $ RationalL r
 derefToExp _ (DerefString s) = LitE $ StringL s
 derefToExp s (DerefList ds) = ListE $ map (derefToExp s) ds
+derefToExp s (DerefTuple ds) = TupE $ map (derefToExp s) ds
 
 -- FIXME shouldn't we use something besides a list here?
 flattenDeref :: Deref -> Maybe [String]
